flatten = lambda l: [item for sublist in l for item in sublist]

u_volume_metric = {
    "cl": [r"cls?", r"centilit(?:er|re)s?"],
    "ml": [r"mls?", r"milliliter", r"millilitre"],
    "l": [r"l", r"liters?", r"litres?"],
    "dl": [r"dls?", r"deciliters?", r"decilitres?"],
}

u_volume_imperial = {
    "tsp": [r"tsps?\.?", r"t", r"tspn\.?", r"(?:heaping )?teaspoons?"],
    "tbsp": [
        r"(?:heaping )?tablespoons?",
        r"T",
        r"t",
        r"tbls?\.?",
        r"tbs\.",
        r"tbs",
        r"tbsps?\.?",
    ],
    "fl oz": [r"fl ?oz", r"fluid ounces?"],
    "gill": [r"gills?", r"gals?"],
    "cup": [r"cups?", r"c\.?"],
    "pint": [r"pints?", r"p", r"pt", r"fl ?pt"],
    "half-pint": [r"half-pints?"],
    "quart": [r"quarts?", r"q", r"qt", r"fl ?qt"],
    "gallon": [r"gallons?", r"gals?"],
    "jigger": [r"jiggers?"],
}

u_weight_metric = {
    "mg": [r"mgs?", r"milligrams?", r"milligrammes?"],
    "g": [r"g", r"grams?", r"grammes?"],
    "kg": [r"kgs?", r"kilograms?", r"kilogrammes?"],
}

u_weight_imperial = {"lb": [r"lbs?", r"pounds?", r"#s?"], "oz": [r"oz", r"ounces?"]}

u_length_metric = {
    "cm": [r"centimeters?", r"cm"],
    "m": [r"met(?:er|re)s?", r"m"],
    "dm": [r"decimet(?:er|re)s?", r"dm"],
    "mm": [r"millimet(?:er|re)s?", r"mm"],
}

u_length_imperial = {
    "inch": [r"inch(?:es)?", r"in", r'"', r"''"],
    "ft": [r"feet", r"foot", r"ft"],
    "yard": [r"yards?", r"yds?"],
}

u_misc = {
    "twist": [r"twists?"],
    "squeeze": [r"squeezes?"],

    "quarter": [r"quarters?"],
    "half": [r"hal(?:f|ves)"],
    "third": [r"thirds?"],
    "part": [r"parts?"],

    "segment": [r"segments?"],
    "wedge": [r"wedges?"],
    "drizzle": [r"drizzles?"],
    "handful": [r"handfuls?"],
    "knob": [r"knobs?"],
    "tub": [r"tubs?"],
    "rasher": [r"rashers?"],
    "can": [r"cans?"],
    "package": [r"packages?", r"pkgs?\.?"],
    "sachet": [r"sachets?"],
    "pack": [r"packs?"],
    "tin": [r"tins?"],
    "rack": [r"racks?"],
    "pinch": [r"pinch(?:es)?"],
    "slice": [r"slices?"],
    "bunch": [r"bunch(?:es)?"],
    "dash": [r"dash(?:es)?"],
    "jar": [r"jars?"],
    "container": [r"containers?", r"cont\.?"],
    "bottle": [r"bottles?"],
    # 'bone': [r'bones?'],
    "chunk": [r"chunks?"],
    "box": [r"box(?:es)?"],
    "stem": [r"stems?"],
    "bulb": [r"bulbs?"],
    "clove": [r"cloves?(?=.*garlic)", r"(?<=garlic )cloves?"],
    "stalk": [r"stalks?"],
    "sprig": [r"sprigs?"],
    "head": [r"heads?"],
    "root": [r"roots?"],
    "loaf": [r"loa(?:f|ves)"],
    "piece": [r"pieces?"],
    "packet": [r"packets?"],
    "drop": [r"drops?"],
    "envelope": [r"envelopes?"],
    "sheet": [r"sheets?"],
    "leaf": [r"lea(?:f|ves)"],
    "bag": [r"bags?"],
    "head": [r"heads?"],
    "carton": [r"cartons?"],
    "stick": [r"sticks?"],
    "link": [r"links?"],
    "scoop": [r"scoops?"],
    "cube": [r"cubes?"],
    "recipe": [r"recipes?"],
    "glass": [r"glass(?:es)?"],
    "handful": [r"handfuls?"],
    "splash": [r"splash(?:es)?"],
    # 'half': [r'(?<!half-and-)hal(?:f|ves)(?!-and-half)'],
    "pod": [r"pods?"],
    "square": [r"squares?"],
    "strip": [r"strips?"],
    "ball": [r"balls?"],

    # "zest": [r"zests?"],
    # "zest and juice": [r"zest? and juices?"],
    # "juice": [r"juices?"],
}

states = [
    r"gold(?:en)?",
    "dark",
    "hearty",
    "rich",
    "homemade",
    "tasty",
    "delicious",
    "yummy",
    "good",
    "cheap",
    "expensive",
    "entire",
    "leftover",
    "quick",
    "easy",
    "simple",
    "fine",
    r"\w+[- ]?sodium",
    "creamy",
    "organic",
    "broken",
    "all-butter",
    "sweet",
    "mild",
    "sharp",
    "firm",
    "soft",
    "aged",
    "thin",
    "thick",
    "lean",
    "stoneground",
    r"\w+[- ]?style",
    r"(?:[0-9]+ [0-9]+/[0-9]+|[0-9]+)[- ]?inch(?:[ -]thick)?",
    r"\w+-sodium",
    r"\w+-fat",
    r"free[- ]range",
    "whole",
    "skin-on",
    "hot",
    "cold",
    "warm",
    "boiling",
    r"\b(?<!\.)(?!0+(?:\.0+)?[ -]?%)(?:\d|[1-9]\d|100)(?:(?<!100)\.\d+)?[ -]?%",
    r"fresh(?:ly)?",
    r"[0-9]+[- ]bone",
    "quality",
    "chopped",
    "ground",
    "sliced",
    "minced",
    "cut",
    "diced",
    "peeled",
    "grated",
    "shredded",
    "drained",
    "dry",
    "coarse",
    "crushed",
    "crusted",
    "dried",
    "cooked",
    "boneless",
    "seeded",
    "beaten",
    "skinless",
    "unsalted",
    "cubed",
    "rinsed",
    "melted",
    "softened",
    "canned",
    "toasted",
    "crumbled",
    "quartered",
    "uncooked",
    "roasted",
    "trimmed",
    "pitted",
    "thawed",
    "granulated",
    "frozen",
    "washed",
    "ripe",
    "mixed",
    "powdered",
    "cored",
    "unsweetened",
    "condensed",
    "smoked",
    "deveined",
    "reduced",
    "mashed",
    "halved",
    "separated",
    "boiled",
    "raw",
    "undrained",
    "seasoned",
    "sifted",
    "slivered",
    "juiced",
    "squeezed",
    "sweetened",
    "evaporated",
    "whipped",
    "blanched",
    "chilled",
    "julienned",
    "cracked",
    "prepared",
    "shelled",
    "stemmed",
    "refried",
    "stewed",
    "unbleached",
    "flaked",
    "skinned",
    "flavored",
    "refrigerated",
    "semisweet",
    "unpeeled",
    "smashed",
    "salted",
    "pickled",
    "packed",
    "snipped",
    "seedless",
    "bottled",
    "deseeded",
    "soaked",
    "rolled",
    "cooled",
    "candied",
    "warmed",
    "baked",
    "scrubbed",
    "steamed",
    "distilled",
    "pressed",
    "corned",
    "marinated",
    "brewed",
    "undiluted",
    "shaved",
    "clarified",
    "stuffed",
    "pureed",
    "unflavored",
    "desiccated",
    "creamed",
    "grilled",
    "zested",
    "hulled",
    "substituted",
    "defrosted",
    "unbaked",
    "assorted",
    "heated",
    "husked",
    "preserved",
    "strained",
    "browned",
    "scraped",
    "thickened",
    "processed",
    "cleaned",
    "jarred",
    "bruised",
    "unseasoned",
    "chunked",
    "crystallized",
    "scalded",
    "rubbed",
    "precooked",
    "buttered",
    "carbonated",
    "stoned",
    "sauteed",
    "tied",
    "shucked",
    "unsifted",
    "butterflied",
    "bone-in",
    "colored",
    "debearded",
    "cured",
    "spiced",
    "spicy",
    "converted",
    "imported",
    "seperated",
    "tinned",
    "blended",
    "added",
    "whisked",
    "deboned",
    "hardboiled",
    "shaped",
    "purchased",
    "fermented",
    "sectioned",
    "veined",
    "suggested",
    "unsliced",
    "pounded",
    "ripened",
    "parboiled",
    "packaged",
    "sorted",
    "textured",
    "dehydrated",
    "grounded",
    "refined",
    "filtered",
    "poached",
    "chipped",
    "compressed",
    "segmented",
    "sieved",
    "heaped",
    "jointed",
    "flattened",
    "combined",
    "unrefined",
    "gutted",
    "caramelized",
    "sprinkled",
    "scaled",
    "pulled",
    "desired",
    "based",
    "rendered",
    "tailed",
    "unblanched",
    "brushed",
    "stripped",
    "diluted",
    "stirred",
    "homogenized",
    "barbecued",
    "concentrated",
    "wedged",
    "herbed",
    "rehydrated",
    "breaded",
    "deviened",
    "scored",
    "rounded",
    "wiped",
    "posted",
    "defatted",
    "filleted",
    "frenched",
    "skin-on",
    "unwaxed",
    "reconstituted",
    "tossed",
    "flavoured",
    "unshelled",
    "choped",
    "tenderized",
    "charred",
    "pasteurized",
    "malted",
    "unsmoked",
    "filled",
]


colors = [
    "red",
    "white",
    "black",
    "brown",
    "green",
    "yellow",
    "blue",
    "purple",
    r"gold(?:en)?",
    "dark",
    r"orange(?=.*pepper|.*beans?)",
]

s_substitutions = {
    "xs": [r"extra small", "mini", "tiny", "super small", "petite"],
    "sm": [
        r"small",
        r"small small",
        r"small[- ]medium",
        r"small[ -]to[ -]medium",
        r"thumb[- ]sized",
    ],
    "md": [r"medium", r"medium small", r"medium[- ]large", r"mediuum", r"medum"],
    "lg": [r"large", r"extra[- ]large", r"large/extra", r"big", r"huge", r"giant", r"massive"],
    "misc": [r"(?:\w|-)+[- ]sized"],
}

u_volume_conversions = {
    "tsp": 5,
    "tbsp": 14.8,
    "cup": 250,
    "fl oz": 30,
    "quart": 946,
    "l": 1000,
    "dl": 100,
    "ml": 1,
    "pint": 473.17,
    "half-pint": 236.585,
    "gallon": 3785.41,
    "jigger": 44,
    "gill": 118.29,
}

u_weight_conversions = {
    "kg": 1000,
    "g": 1,
    "mg": 0.001,
    "oz": 28.34,
    "lb": 453.59,
}

n_substitutions = {
    None: [r"(?:a )?few", r"a little", r"some", r"several"],
    24: [r"(?:two|2) dozens?"],
    36: [r"(?:three|3) dozens?"],
    12: [r"dozen", r"(?:one|1) dozen", r"a dozen"],
    1: [r"one", r"a", r"an"],
    2: [r"two"],
    3: [r"three"],
    4: [r"four"],
    5: [r"five"],
    6: [r"six"],
    7: [r"seven"],
    8: [r"eight"],
    9: [r"nine"],
    10: [r"ten"],
    0.5: [r"half"],
}


u_misc_values = flatten(u_misc.values())
u_all_values = flatten(
    [
        flatten(list(unit_dict.values()))
        for unit_dict in [
            u_volume_imperial,
            u_volume_metric,
            u_misc,
            u_weight_imperial,
            u_weight_metric,
            u_length_imperial,
            u_length_metric,
        ]
    ]
)
u_metric_values = flatten(
    [
        flatten(list(unit_dict.values()))
        for unit_dict in [u_volume_metric, u_weight_metric, u_length_metric]
    ]
)

u_imperial_values = flatten(
    [
        flatten(list(unit_dict.values()))
        for unit_dict in [u_volume_imperial, u_weight_imperial, u_length_metric]
    ]
)

word_numbers = flatten(n_substitutions.values())

u_substitutions = {}
for set in [
    u_weight_imperial,
    u_weight_metric,
    u_volume_imperial,
    u_volume_metric,
    u_misc,
    u_length_imperial,
    u_length_metric,
]:
    for key, values in set.items():
        u_substitutions[key] = rf"(?:\b(?:{r'|'.join(values)})\b)"


sizes = flatten(s_substitutions.values())



from .data.constants import (
    u_all_values,
    u_misc_values,
    sizes,
    word_numbers,
    states,
    u_imperial_values,
    u_metric_values,
    colors,
)
from .data import meat
# import re
# from pprint import pprint

range_separators = [r"-", r"to", r"–", r"or"]


def one_of(l):
    return r"|".join(l)


L_B = r"(?<!\w)"
R_B = r"(?!=\w)"
N_B = r"(?<=\.|,|/| ?-)(?=\.|,|/| ?-)"


N_PREP = rf"{L_B}(?:each of(?= )|of each(?= )|of an?(?= )|of(?= )|an?(?= )|each,(?= )|each(?= ))"
N_WORD = rf"\b(?:{one_of(word_numbers)})\b"
N_FRACTION = r"\d+ ?/ ?\d+"
N_DECIMAL = r"\d+\.\d+"
N_WHOLE = r"\d+(?![/\.])"
N_COMPOSED = rf"{N_WHOLE} {N_FRACTION}"
NUMBER = (
    rf"(?:"
    rf"{N_WORD}(?= |$)"
    rf"|"
    rf"{N_DECIMAL}"
    rf"|"
    rf"{N_FRACTION}"
    rf"|"
    rf"{N_WHOLE}(?: {N_FRACTION})?"
    rf")"
)


R_SEP = rf" ?(?:{one_of(range_separators)}) ?"
RANGE = rf"(?:" rf"{NUMBER}" rf"{R_SEP}" rf"{NUMBER}" rf")"


MOD = (
    rf"(?:"
    rf"\b\w+[- ]?(?:less|free|ful)\b"
    rf"|"
    rf"\b(?:\w+ly )?(?:(?:\w|-)+-|un|pre|extra|over|de)?(?:{one_of(states)})(?:-[\w-]+)?\b"
    rf"|"
    rf"\b(?:low|reduced|high|non)[- ]?(?:\w+)\b"
    rf")"
)
MOD_SEP = r"(?:(?: ?,)? or | (?: ?,)? and/or | ?, ?|(?: ?,)? and | ?& ?| to | )"
MODS = rf"(?:{MOD}{MOD_SEP})*{MOD}"

COLOR = rf"(?:" rf"{L_B}(?:{one_of(colors)}){R_B}" rf")"
COLORS = rf"(?:{COLOR}{MOD_SEP})*{COLOR}"

SIZE = rf"(?:{L_B}(?:{one_of(sizes)})(?:[- ]sized)?{R_B})"
SIZES = rf"(?:{SIZE}{MOD_SEP})*{SIZE}"


U = r"(?:(?<![a-z])(?:" + one_of(u_all_values) + r")(?![a-z]))"
U_IMPERIAL = r"(?:(?<![a-z])(?:" + one_of(u_imperial_values) + r")(?![a-z]))"
U_METRIC = r"(?:(?<![a-z])(?:" + one_of(u_metric_values) + r")(?![a-z]))"
U_MISC = (
    rf"(?:(?:{N_PREP} )?(?:{SIZES} {MOD_SEP}?)?(?:{MODS} )?(?<![a-z])(?:"
    + one_of(u_misc_values)
    + r")(?![a-z–-]))"
)
UNIT = rf"(?:(?:(?:generous|heaping|heaped) )?{U}\.?)"

Q = rf"(?:{RANGE}|{NUMBER})(?!-)"
Q_UNIT = rf"(?:{Q}[ -]?{UNIT})"
Q_SIZE = rf"(?:{Q} {SIZE})"

# Q_COMPOSED = rf'(?:{Q_UNIT} {Q_UNIT})'
Q_COMPOSED = (
    rf"(?:"
    rf"{Q} ?{U_IMPERIAL}\.? ?{Q} ?{U_IMPERIAL}\.?"
    rf"|"
    rf"{Q} ?{U_METRIC}\.? ?{Q} ?{U_METRIC}\.?"
    rf")"
)


# Q_OPTIONS = rf'(?:{Q_COMPOSED}|{Q_UNIT}|{Q_SIZE}) ?(?:\/|or) ?(?:{Q_COMPOSED}|{Q_UNIT}|{Q_SIZE})'
Q_DIMS = (
    rf"(?:"
    rf"(?:{NUMBER}[- –]?{UNIT})"
    rf" ?(?:by|x) ?"
    rf"(?:{NUMBER}[- –]?{UNIT})"
    rf")"
)
Q_UNIT_MISC = (
    rf"(?:"
    rf"(?:(?P<multiplier>{Q} )?(?P<quantity>{Q})-(?P<unit>{UNIT}))"
    rf" (?P<misc>{U_MISC})"
    rf")"
)

MULTIPLIER = rf"(?:(?P<multiplier>{Q}) ?[x*] ?)"

Q_UNIT_RANGE = (
    rf"(?:" rf"{Q_COMPOSED}{R_SEP}{Q_COMPOSED}" rf"|" rf"{Q_UNIT}{R_SEP}{Q_UNIT}" rf")"
)

QUANTITY = (
    rf"(?:"
    rf"(?:(?P<multiplier>{Q}) ?[x*] ?)?"
    rf"(?P<quantity>{Q_DIMS}|{Q_UNIT_RANGE}|{Q_COMPOSED}|{Q_UNIT}|{Q}|{UNIT})?"
    rf"(?:"
    rf"(?: ?/ ?| or )"
    rf"(?:(?P<multiplier_alt>{Q}) ?[x*] ?)?"
    rf"(?P<quantity_alt>{Q_DIMS}|{Q_UNIT_RANGE}|{Q_COMPOSED}|{Q_UNIT}|{Q}|{UNIT})?"
    rf")?"
    rf"(?: ?(?P<misc_size>{SIZES})?{MOD_SEP}?(?P<misc_mods>{MODS})? ?(?P<extra_misc>{U_MISC}))?"
    rf")"
)


START_IRREG = rf"(?:sliced |cut |\w+ into |(?:\w+ed )?in |with(?:out)? |from |- |on |to (?=[a-z])|for |weighing |mixed (?:into |with |to |in )|is )"
START_POST_MOD = rf" ?, ?"
START_ALT_ING = rf"(?: or | and | & )"
POST_UNIT = rf"(?:(?:{one_of(u_misc_values)}\b)(?=$|{START_POST_MOD}|{START_ALT_ING}| {START_IRREG}))"

END_ING = rf"$| {START_IRREG}|{START_POST_MOD}| {POST_UNIT}|{START_ALT_ING}"

IRREG_POST_MODS = (
    rf"(?P<irreg_post_mod>" rf"{START_IRREG}[^,]*?(?= ?, ?| or | and |$)" rf")"
)

INGREDIENT = (
    rf"(?:about |approx(?:imately)? |around |plus |additional |extra |to taste |more )?"
    rf"(?:{QUANTITY} ?)?"
    rf"(?:or so |or about |about |approximately |around |per person )?"
    rf"(?:{N_PREP} )?"
    rf"(?:(?P<size>{SIZES}){MOD_SEP})?"
    rf"(?:(?P<pre_mod>{MODS}){MOD_SEP})?"
    rf"(?:(?P<color>{COLORS}) )?"
    rf"(?P<rest>"
    rf"(?:(?P<ingredient>.+?(?={END_ING})))?"
    rf"(?: ?(?P<post_unit>{POST_UNIT}))?"
    rf"(?: ?{IRREG_POST_MODS})?"
    rf"(?: ?, (?P<post_mod>.+?(?= or | and | & |$)))?"
    rf"(?:(?: or | and | & )(?P<ingredient_alt>{Q}.*))?"
    rf")"
)

PROTEIN = one_of(
    [
        rf"""(?P<{key.replace(' ', '_')}>{one_of(values)})(?:$|.*(?P<{key.replace(' ', '_')}_portion>{
                one_of(meat.portions[
                    meat.type_portions[key] if key in meat.type_portions else 'all'
                ] + meat.portions['general'])
            }))"""
        for key, values in meat.dictionary.items()
    ]
)

tests = [
    "175g/6oz piece smoked pancetta , rind removed",
    "1 inch x 2 inch large knob ginger",
    "1/4 cup celery , finely chopped",
    "3-4 lbs beef round steak",
    "1 1/2 head celery",
    "1kg 50g beef",
    "100ml/3 1/2fl oz double cream",
    "1 1/2-2kg/3lb 5oz - 4lb 8oz lamb shoulder",
    "5 1/2 cloves garlic , crushed",
    "5 tsp cloves",
    "5 garlic cloves to taste, crushed",
    "3 3-ounce bags of chilli powder, smoked whoole",
    "1 x 6-bone rack of lamb",
    "1 1/2 big 6-bone racks of lamb",
    "1 big rack of lamb or 2 megapints",
    "freshly ground black pepper to taste",
    "3 bay leaves",
    "7-8 cinnamon sticks",
    "1 thick crusted baguette",
    "small bunch tarragon , roughly chopped",
    "3 handfuls fresh, raw, small shelled prawns",
]

# for test in tests:
#     pprint(test)
#     pprint(re.search(INGREDIENT, test).groupdict())

# print(
#     re.search(INGREDIENT, "freshly ground black pepper to taste").groupdict()
# )



from ..utils import *
import html
from .. import regex as re_
from ..data.ingredient_funnel import dictionary as ing_funnel
import numpy as np

class RecipeReader:
    def __init__(self):
        pass

    def read_phrase(self, phrase):
        if not P_filter(phrase):
            return None

        phrase = html.unescape(phrase)
        phrase = normalizer.normalize_str(phrase)
        phrase = P_duplicates(phrase)

        phrase = P_multi_misc_fix(phrase)
        phrase = P_multi_misc_fix(phrase)
        phrase = P_missing_multiplier_symbol_fix(phrase)
        phrase = P_quantity_dash_unit_fix(phrase)
        phrase = P_juice_zest_fix(phrase)

        values = re.search(re_.INGREDIENT, phrase).groupdict()

        if values["quantity"]:
            values["quantity"], values["unit"] = re.search(
                rf"(?P<quantity>{re_.Q})? ?(?P<unit>.*)?", values["quantity"]
            ).groups()
            values["quantity"] = Q_to_number(values["quantity"])
        values["unit"] = None
        values["unit"] = U_unify(values["unit"])
        values["size"] = S_unify(values["size"])

        if values["ingredient"] != values["ingredient"]:
            return None

        values["ingredient"] = I_to_singular(values["ingredient"])
        values["simple"] = I_label_protein(values["ingredient"])
        values["simple"] = I_simplify(values["simple"])

        filtered = {
            c: values[c]
            for c in ["quantity", "unit", "size", "color", "ingredient", "simple"]
        }
        filtered["simple"] = values["simple"]
        return values["simple"]


    def funnel(self, phrase):
        return ing_funnel[phrase] if phrase in ing_funnel else None


    def read(self, title, phrases, steps):
        out = [self.read_phrase(p) for p in phrases]
        columns = sorted(list(set(list(ing_funnel.values()))))
        out = [self.funnel(p) for p in out if self.funnel(p)]

        return {
            'ingredients': out,
            'values': np.array([1 if c in out else 0 for c in columns])
        }